@@ -145,8 +145,7 @@ class Session:
         self.extended_smtp = False
         self.loop = loop
 
-        self.login_data = None
-        """Legacy login_data, usually containing the username"""
+        self._login_data = None
 
         self.auth_data = None
         """
@@ -158,6 +157,21 @@ class Session:
 
         self.authenticated = None
 
+    @property
+    def login_data(self):
+        """Legacy login_data, usually containing the username"""
+        log.warning(
+            "Session.login_data is deprecated and will be removed in version 2.0",
+        )
+        return self._login_data
+
+    @login_data.setter
+    def login_data(self, value):
+        log.warning(
+            "Session.login_data is deprecated and will be removed in version 2.0",
+        )
+        self._login_data = value
+
 
 @public
 class Envelope:
